from .ami import get_by_tag
import boto3
import six
import os.path
from botocore.configloader import raw_config_parse


class Config:
    def __init__(self, client, args, config_file_path = "~/.spotr/config"):
        self.client = client
        if os.path.isfile(config_file_path):
            self._config = raw_config_parse(config_file_path)['config']
        else:
            self._config = {}
        self._config.update({k: v for k, v in six.iteritems(vars(args)) if v})

    def set_az(self, az):
        self._config['az'] = az

    @property
    def ami_tag(self):
        return self._config.get('ami_tag', 'spotr')

    @property
    def instance_tag(self):
        return self._config.get('instance_tag', 'spotr')

    @property
    def type(self):
        return self._get_required('type')

    @property
    def max_bid(self):
        return self._get_required('max_bid')

    @property
    def ami(self):
        if 'ami' not in self._config:
            self._config['ami'] = get_by_tag(self.client, self.ami_tag)
        return self._config['ami']

    @property
    def key_name(self):
        return self._config.get('key_name', 'spotr')

    @property
    def az(self):
        return self._config['az']


    def _get_required(self, key):
        if not self._config.get(key):
            raise RuntimeError("Missing required parameter: {0}".format(key))
        return self._config.get(key)



import threading
import sys, os
import time
import unicodedata
from contextlib import contextmanager

@contextmanager
def spin(message):
    spin = SpinCursor(msg=message)
    spin.start()
    try:
        yield
    finally:
        spin.stop()


class SpinCursor(threading.Thread):
    """ A console spin cursor class """
    
    def __init__(self, msg='',maxspin=0,minspin=10,speed=5):
        # Count of a spin
        self.count = 0
        self.out = sys.stdout
        self.flag = False
        self.max = maxspin
        self.min = minspin
        # Any message to print first ?
        self.msg = msg
        # Complete printed string
        self.string = ''
        # Speed is given as number of spins a second
        # Use it to calculate spin wait time
        self.waittime = 1.0/float(speed*4)
        if os.name == 'posix':
            self.spinchars = (unicodedata.lookup('FIGURE DASH'),u'\\ ',u'| ',u'/ ')
        else:
            # The unicode dash character does not show
            # up properly in Windows console.
            self.spinchars = (u'-',u'\\ ',u'| ',u'/ ')        
        threading.Thread.__init__(self, None, None, "Spin Thread")
        
    def spin(self):
        """ Perform a single spin """

        for x in self.spinchars:
            self.string = self.msg + "...\t" + x + "\r"
            self.out.write(self.string)
            self.out.flush()
            time.sleep(self.waittime)

    def run(self):

        while (not self.flag):
            self.spin()
            self.count += 1

        # Clean up display...
        self.out.write(" "*(len(self.string) + 1))
        
    def stop(self):
        self.flag = True
        if self.is_alive():
            self.join()



class Instance():
    def __init__(self, response):
        self.id = response['InstanceId']
        self.volume_id = response['BlockDeviceMappings'][0]['Ebs']['VolumeId']
        self.launch_time = response['LaunchTime']
        self.ip_address = response['PublicIpAddress']
        self.security_groups = response['SecurityGroups']

    @property
    def has_security_groups(self):
        return not len(self.security_groups) == 0


class InstanceList():
    def __init__(self, response):
        self.instances = []
        for reservation in response['Reservations']:
            for insta in reservation['Instances']:
                if 'PublicIpAddress' in insta:
                    self.instances.append(Instance(insta))

    def latest(self):
        return sorted(self.instances, key=lambda i: i.launch_time)[0]

    def __len__(self):
        return len(self.instances)


def find_latest(client, config):
    instance_response = client.describe_instances(
        Filters=[{'Name': 'tag:project', 'Values': [config.instance_tag]}])
    instance_list = InstanceList(instance_response)

    if len(instance_list) == 0:
        raise RuntimeError("No running spotr instances found")

    return instance_list.latest()


def destroy(client, instance_id):
    return client.terminate_instances(InstanceIds=[instance_id])


def tag(client, instance_id, config):
    return client.create_tags(
        Resources=[instance_id],
        Tags=[{'Key': 'project', 'Value': config.instance_tag, }]
    )


def get_by_instance_id(client, instance_id):
    response = client.describe_instances(InstanceIds=[instance_id])
    return Instance(response['Reservations'][0]['Instances'][0])

def open_port(client, instance, port):
    if len(instance.security_groups) == 0:
        raise RuntimeError("No security groups associated with instance")
    group_id = instance.security_groups[0].get('GroupId')
    group = client.describe_security_groups(GroupIds=[group_id])['SecurityGroups'][0]
    matching_rules = (x for x in group['IpPermissions'] if x.get('FromPort') == port and x.get('ToPort') == port)
    first_matching_rule = next(matching_rules, None)
    if not first_matching_rule:
        client.authorize_security_group_ingress(
            GroupId = group_id,
            IpPermissions=[
                {
                    'FromPort': port,
                    'IpProtocol': 'TCP',
                    'IpRanges': [
                        {
                            'CidrIp': '0.0.0.0/0',
                            'Description': 'Spotr Jupyter Port'
                        },
                    ],
                    'Ipv6Ranges': [
                        {
                            'CidrIpv6': '::/0',
                            'Description': 'Spotr Jupyter Port'
                        },
                    ],
                    'ToPort': port
                }
            ]
        )
    return True



from .pricing import get_az
from .spot_instance import request
from .instance import tag as tag_instance
from .instance import get_by_instance_id
from .instance import open_port
from .client import build as build_client
from .config import Config
from .key import find_or_create as find_or_create_key
from .spin_cursor import spin


def launch(args):
    client = build_client(args)
    conf = Config(client, args)

    key_path = find_or_create_key(client, conf.key_name)

    az = get_az(client, conf)
    conf.set_az(az.zone_name)

    with spin("Launching: " + str(az)):
        inst = request(client, conf, tag_instance, get_by_instance_id, open_port)

    _log_instance_creation(inst, key_path)

    return inst


def _log_instance_creation(instance, key_path):
    print(">> Instance launched, connect with:")
    ip = str(instance.ip_address)
    print("ssh -i " + key_path + " ubuntu@" + ip)



def get_by_tag(client, tag):
    response = client.describe_images(
        Filters=[{'Name': 'tag:project', 'Values': [tag]}]
    )
    images = response['Images']
    if len(images) == 0:
        raise RuntimeError("No saved snapshots with tag: '{0}'".format(tag))
    return images[0]['ImageId']



import datetime
import sys
import boto3

from .availability_zone import AvailabilityZone


def get_az(client, config):
    zone_names = _get_zone_names(client)
    instance_type = config.type

    azs = []
    for zone_name in zone_names:
        price_history = _get_price_history(client, zone_name, instance_type)
        az = AvailabilityZone(zone_name, price_history)
        azs.append(az)

    return sorted(azs, key=_score)[0]


def _get_zone_names(client):
    zone_names = []
    for zone in client.describe_availability_zones()['AvailabilityZones']:
        if zone['State'] == 'available':
            zone_names.append(zone['ZoneName'])
    return zone_names


def _get_price_history(client, zone_name, instance_type):
    response = client.describe_spot_price_history(
        DryRun=False,
        StartTime=datetime.datetime.now() - datetime.timedelta(days=7),
        EndTime=datetime.datetime.now(),
        InstanceTypes=[instance_type],
        AvailabilityZone=zone_name,
        ProductDescriptions=['Linux/UNIX'])
    return response.get('SpotPriceHistory', [])


def _score(az):
    return az.current_price or sys.maxsize



import argparse
import sys

from .launch import launch
from .destroy import destroy
from .snapshot import snapshot
from .version import VERSION


parser = argparse.ArgumentParser()
parser.add_argument('--version', action='version', version=VERSION)
subparsers = parser.add_subparsers()

launch_parser = subparsers.add_parser('launch')
launch_parser.add_argument(
    '--type',
    help='the type of the instance to launch, eg p2.xlarge2')
launch_parser.add_argument(
    '--region',
    help='the region to launch the instance in')
launch_parser.add_argument(
    '--aws-access-key-id',
    help='the access key id to use')
launch_parser.add_argument(
    '--aws-secret-access-key',
    help='the secret access key to use')
launch_parser.add_argument(
    '--max-bid',
    help='the maximum bid price')
launch_parser.add_argument(
    '--ami',
    help='id of ami to use')
launch_parser.add_argument(
    '--ami-tag',
    help='the tag to search for the AMI by')
launch_parser.add_argument(
    '--key-name',
    help='name of the aws key pair to use')
launch_parser.set_defaults(func=launch)

destroy_parser = subparsers.add_parser('snapshot')
destroy_parser.add_argument(
    '--region',
    help='the region of the launched spotr instance')
destroy_parser.add_argument(
    '--aws-access-key-id',
    help='the access key id to use')
destroy_parser.add_argument(
    '--aws-secret-access-key',
    help='the secret access key to use')
destroy_parser.set_defaults(func=snapshot)

destroy_parser = subparsers.add_parser('destroy')
destroy_parser.add_argument(
    '--region',
    help='the region of the launched spotr instance')
destroy_parser.add_argument(
    '--aws-access-key-id',
    help='the access key id to use')
destroy_parser.add_argument(
    '--aws-secret-access-key',
    help='the secret access key to use')
destroy_parser.set_defaults(func=destroy)


def main():
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



import time
from .instance import find_latest as find_latest_instance
from .instance import destroy as destroy_instance
from .client import build as build_client
from .config import Config
from .spin_cursor import spin


def destroy(args):
    client = build_client(args)
    conf = Config(client, args)
    instance = find_latest_instance(client, conf)

    with spin(">> Destroying instance."):
        destroyed = destroy_instance(client, instance.id)

    _log_instance_destroyed(instance)

    return destroyed

def _log_instance_destroyed(instance):
    print(">> Instance {0} destroyed".format(instance.id))




class AvailabilityZone:
    def __init__(self, zone_name, price_history):
        self.zone_name = zone_name
        self.price_history = price_history

    @property
    def current_price(self):
        if self.price_history:
            return float(self.price_history[0]['SpotPrice'])
        else:
            return None

    def __repr__(self):
        price = str(self.current_price)
        zone_name = self.zone_name
        return "%s for $%s/hr" % (zone_name, price)



import boto3
import time
from .client import build as build_client
from .instance import find_latest as find_latest_instance
from .config import Config
from .spin_cursor import spin


def snapshot(args):
    client = build_client(args)
    conf = Config(client, args)
    instance = find_latest_instance(client, conf)

    with spin(">> Creating snapshot for instance: " + str(instance.ip_address)):
        snap = create_and_wait(client, instance.volume_id)
    return snap

def create_and_wait(client, volume_id):
    snapshot = create(client, volume_id)
    wait_for_completion(client, snapshot)
    return snapshot


def create(client, volume_id):
    response = client.create_snapshot(
        Description="Spotr snapshot",
        VolumeId=volume_id)
    return Snapshot(response)


def wait_for_completion(client, snapshot):
    waiter = client.get_waiter('snapshot_completed')
    waiter.wait(
        Filters=[
            {
                'Name': 'snapshot-id',
                'Values': [snapshot.id]
            },
        ]
    )

class Snapshot():
    def __init__(this, response):
        this.state = response['State']
        this.id = response['SnapshotId']

